# Modified from:
# https://gist.github.com/flatz/3f242ab3c550d361f8c6d031b07fb6b1

import json
import subprocess
import sys
from pathlib import Path
from typing import Optional

if sys.platform == "win32":
    from base64 import b64decode
    from ctypes import *
    from ctypes.wintypes import DWORD

    class DataBlob(Structure):
        _fields_ = [("cbData", DWORD), ("pbData", POINTER(c_char))]


from Crypto.Cipher import AES
from Crypto.Hash import SHA1
from Crypto.Protocol.KDF import PBKDF2
from Crypto.Util.Padding import unpad
from typer import colors, secho

PASSWORD_CMD_DARWIN = [
    "security",
    "find-generic-password",
    "-ws",
    "Signal Safe Storage",
]
PASSWORD_CMD_GNOME = ["secret-tool", "lookup", "application", "Signal"]
PASSWORD_CMD_KDE = [
    "kwallet-query",
    "kdewallet",
    "-f",
    "Chromium Keys",
    "-r",
    "Chromium Safe Storage",
]


def get_key(appdir: Path, password: Optional[str]) -> Optional[str]:
    """Get key for decrypting database.

    Retrieves key depending on key encryption software.

    If it cannot be decrypted, print an explanation message.

    Args:
        file: Signal config json file path
        password: password that user could have supplied to decrypt key
    Returns:
        (decrypted) password or None
    """

    with open(appdir / "config.json", encoding="utf-8") as cf:
        data = json.loads(cf.read())
    if "key" in data:
        return data["key"]
    elif "encryptedKey" in data:
        encrypted_key = data["encryptedKey"]
        if sys.platform == "win32":
            if not password:
                with open(appdir / "Local State", encoding="utf-8") as lsf:
                    data = json.loads(lsf.read())
                if "os_crypt" in data and "encrypted_key" in data["os_crypt"]:
                    pw_encrypted_b64 = data["os_crypt"]["encrypted_key"]
                else:
                    secho("Encrypted password not found in Local State", fg=colors.RED)
                    raise

                # base64decode the encrypted password, and cut off the first 5 bytes ('D' 'P' 'A' 'P' 'I')
                pw_encrypted = b64decode(pw_encrypted_b64)[5:]

                # decrypt the password
                data_in = DataBlob(
                    len(pw_encrypted), c_buffer(pw_encrypted, len(pw_encrypted))
                )
                data_out = DataBlob()
                if windll.crypt32.CryptUnprotectData(
                    byref(data_in), None, None, None, None, 0, byref(data_out)
                ):
                    cbData = int(data_out.cbData)
                    pbData = data_out.pbData
                    buffer = c_buffer(cbData)
                    cdll.msvcrt.memcpy(buffer, pbData, cbData)
                    windll.kernel32.LocalFree(pbData)
                    pw = buffer.raw
                else:
                    secho("Failed to decrypt password", fg=colors.RED)
                    raise
            else:
                pw = bytearray.fromhex(password)

            # The encrypted key consists of the following parts:
            # 3 bytes header ('V' '1' '0')
            # 12 bytes nonce
            # 64 bytes encrypted data
            # 16 bytes MAC
            encryptedKey_struct = memoryview(bytearray.fromhex(encrypted_key))
            key = AES.new(
                pw, AES.MODE_GCM, nonce=encryptedKey_struct[3:15]
            ).decrypt_and_verify(encryptedKey_struct[15:79], encryptedKey_struct[79:])
            return key.decode("ascii")
        if sys.platform == "darwin":
            if password:
                return decrypt(password, encrypted_key, b"v10", 1003)
            pw = get_password(PASSWORD_CMD_DARWIN, "macOS")  # may raise error
            return decrypt(pw, encrypted_key, b"v10", 1003)
        else:  # linux
            if password:
                return decrypt(password, encrypted_key, b"v11", 1)
            elif "safeStorageBackend" in data:
                if data["safeStorageBackend"] == "gnome_libsecret":
                    pw = get_password(PASSWORD_CMD_GNOME, "gnome")  # may raise error
                    return decrypt(pw, encrypted_key, b"v11", 1)
                elif data["safeStorageBackend"] in [
                    "gnome_libsecret",
                    "kwallet",
                    "kwallet5",
                    "kwallet6",
                ]:
                    pw = get_password(PASSWORD_CMD_KDE, "KDE")  # may raise error
                    return decrypt(pw, encrypted_key, b"v11", 1)
                else:
                    secho("Your Signal data key is encrypted, and requires a password.")
                    secho(f"The safe storage backend is {data['safeStorageBackend']}")
                    secho(
                        "If you know some Python and know how to retrieve passwords "
                        "from this backend, please contribute a PR!"
                    )
            else:
                secho("Your Signal data key is encrypted, and requires a password.")
                secho("No safe storage backend is specified.")
                secho(
                    "On gnome, you can usually retrieve the password with the command"
                )
                secho(" ".join(PASSWORD_CMD_GNOME) + "\n", fg=colors.BLUE)
                secho("On KDE, you can usually retreive the password with the command")
                secho(" ".join(PASSWORD_CMD_KDE) + "\n", fg=colors.BLUE)
                secho(
                    "If you have found your password, please rerun sigexport as follows:"
                )
                secho("sigexport --password=PASSWORD_FROM_COMMAND ...", fg=colors.BLUE)
                secho("No Signal decryption key found", fg=colors.RED)
    else:
        secho("No Signal decryption key found", fg=colors.RED)


def get_password(cmd: list[str], system: str) -> str:
    """Call external tool to get key password.

    Args:
        cmd: shell command as list of words
        system: Name of the system we are on, for help message.
    Returns:
        password if found
    Raises:
        nondescript error: if no password was found
    """
    try:
        p = subprocess.run(  # NoQA: S603
            cmd, capture_output=True, text=True, encoding="utf-8"
        )
    except FileNotFoundError as e:
        secho(
            f"When trying to retrieve the password, '{e.filename}' was not found. "
            "You may need to install the respective package."
        )
        raise
    if p.returncode != 0:
        secho("Your Signal data key is encrypted, and requires a password.")
        secho(f"Usually on {system}, you can try to get it with this command:")
        secho(" ".join(cmd) + "\n", fg=colors.BLUE)
        secho(
            "But this failed with errorcode "
            f"{p.returncode} and error {p.stdout} {p.stderr}"
        )
        secho("If you have found your password, please rerun sigexport as follows:")
        secho("sigexport --password=PASSWORD_FROM_COMMAND ...", fg=colors.BLUE)
        secho("No Signal decryption key found", fg=colors.RED)
        raise
    pw = p.stdout
    return pw.strip()


def decrypt(password: str, encrypted_key: str, prefix: bytes, iterations: int) -> str:
    encrypted_key_b = bytes.fromhex(encrypted_key)
    if not encrypted_key_b.startswith(prefix):
        raise
    encrypted_key_b = encrypted_key_b[len(prefix) :]

    salt = b"saltysalt"
    key = PBKDF2(
        password, salt=salt, dkLen=128 // 8, count=iterations, hmac_hash_module=SHA1
    )
    iv = b" " * 16
    aes_decrypted = AES.new(key, AES.MODE_CBC, iv).decrypt(encrypted_key_b)
    decrypted_key = unpad(aes_decrypted, block_size=16).decode("ascii")
    return decrypted_key



import base64
import hashlib
import hmac
import shutil
from datetime import datetime
from pathlib import Path

from Crypto.Cipher import AES
from typer import colors, secho

from sigexport import models
from sigexport.logging import log

CIPHER_KEY_SIZE = 32
IV_SIZE = AES.block_size
MAC_KEY_SIZE = 32
MAC_SIZE = hashlib.sha256().digest_size


def decrypt_attachment(att: dict[str, str], src_path: Path, dst_path: Path) -> None:
    """Decrypt attachment and save to `dst_path`.

    Code adapted from:
        https://github.com/tbvdm/sigtop
    """
    try:
        keys = base64.b64decode(att["localKey"])
    except KeyError:
        raise ValueError("No key in attachment")
    except Exception as e:
        raise ValueError(f"Cannot decode keys: {str(e)}")

    if len(keys) != CIPHER_KEY_SIZE + MAC_KEY_SIZE:
        raise ValueError("Invalid keys length")

    cipher_key = keys[:CIPHER_KEY_SIZE]
    mac_key = keys[CIPHER_KEY_SIZE:]

    try:
        with open(src_path, "rb") as fp:
            data = fp.read()
    except Exception as e:
        raise ValueError(f"Failed to read file: {str(e)}")

    if len(data) < IV_SIZE + MAC_SIZE:
        raise ValueError("Attachment data too short")

    iv = data[:IV_SIZE]
    their_mac = data[-MAC_SIZE:]
    data = data[IV_SIZE:-MAC_SIZE]

    if len(data) % AES.block_size != 0:
        raise ValueError("Invalid attachment data length")

    m = hmac.new(mac_key, iv + data, hashlib.sha256)
    our_mac = m.digest()

    if not hmac.compare_digest(our_mac, their_mac):
        raise ValueError("MAC mismatch")

    try:
        cipher = AES.new(cipher_key, AES.MODE_CBC, iv)
        decrypted_data = cipher.decrypt(data)
    except Exception as e:
        raise ValueError(f"Decryption failed: {str(e)}")

    if len(decrypted_data) < int(att["size"]):
        raise ValueError("Invalid attachment data length")

    data_decrypted = decrypted_data[: att["size"]]
    with open(dst_path, "wb") as fp:
        fp.write(data_decrypted)


def copy_attachments(
    src: Path, dest: Path, convos: models.Convos, contacts: models.Contacts
) -> None:
    """Copy attachments and reorganise in destination directory."""
    src_root = Path(src) / "attachments.noindex"
    dest = Path(dest)

    for key, messages in convos.items():
        name = contacts[key].name
        log(f"\tCopying attachments for: {name}")
        # some contact names are None
        if not name:
            name = "None"
        dst_root = dest / name / "media"
        dst_root.mkdir(exist_ok=True, parents=True)
        for msg in messages:
            if msg.attachments:
                attachments = msg.attachments
                date = (
                    datetime.fromtimestamp(msg.get_ts() / 1000)
                    .isoformat(timespec="milliseconds")
                    .replace(":", "-")
                )
                for i, att in enumerate(attachments):
                    # Account for no fileName key
                    file_name = str(att["fileName"]) if "fileName" in att else "None"
                    # Sometimes the key is there but it is None, needs extension
                    if "." not in file_name:
                        content_type = att.get("contentType", "").split("/")
                        if len(content_type) > 1:
                            ext = content_type[1]
                        else:
                            ext = content_type[0]
                        file_name += "." + ext
                    att["fileName"] = (
                        f"{date}_{i:02}_{file_name}".replace(" ", "_")
                        .replace("/", "-")
                        .replace(",", "")
                        .replace(":", "-")
                        .replace("|", "-")
                    )
                    # account for erroneous backslash in path
                    try:
                        att_path = str(att["path"]).replace("\\", "/")
                    except KeyError:
                        log(f"\t\tBroken attachment:\t{name}")
                        continue
                    src_path = src_root / att_path
                    dst_path = dst_root / att["fileName"]
                    if int(att.get("version", 0)) >= 2:
                        try:
                            decrypt_attachment(att, src_path, dst_path)
                        except ValueError as e:
                            secho(
                                f"Failed to decrypt {src_path} error {e}, skipping",
                                fg=colors.MAGENTA,
                            )
                    else:
                        try:
                            shutil.copy2(src_path, dst_path)
                        except FileNotFoundError:
                            secho(
                                f"No file to copy at {src_path}, skipping!",
                                fg=colors.MAGENTA,
                            )
                        except OSError as exc:
                            secho(
                                f"Error copying file {src_path}, skipping!\n{exc}",
                                fg=colors.MAGENTA,
                            )
            else:
                msg.attachments = []


def merge_attachments(media_new: Path, media_old: Path) -> None:
    """Merge new and old attachments directories."""
    for f in media_old.iterdir():
        if f.is_file():
            try:
                shutil.copy2(f, media_new)
            except shutil.SameFileError:
                log(
                    f"Skipped file {f} as duplicate found in new export directory!",
                    fg=colors.RED,
                )



"""Main script for sigexport."""

import shutil
from pathlib import Path
from typing import Optional

from typer import Argument, Context, Exit, Option, colors, run, secho

from sigexport import create, data, files, html, logging, merge, utils

OptionalPath = Optional[Path]
OptionalStr = Optional[str]


def main(
    ctx: Context,
    dest: Path = Argument(None),
    source: OptionalPath = Option(None, help="Path to Signal source directory"),
    old: OptionalPath = Option(None, help="Path to previous export to merge"),
    password: OptionalStr = Option(None, help="Linux-only. Password to decrypt DB key"),
    key: OptionalStr = Option(
        None, help="Linux-only. DB key, as found in the old config.json"
    ),
    paginate: int = Option(
        100, "--paginate", "-p", help="Messages per page in HTML; set to 0 for infinite"
    ),
    chats: str = Option(
        "", help="Comma-separated chat names to include: contact names or group names"
    ),
    json_output: bool = Option(
        True, "--json/--no-json", "-j", help="Whether to create JSON output"
    ),
    html_output: bool = Option(
        True, "--html/--no-html", "-h", help="Whether to create HTML output"
    ),
    list_chats: bool = Option(
        False, "--list-chats", "-l", help="List available chats and exit"
    ),
    include_empty: bool = Option(
        False, "--include-empty", help="Whether to include empty chats"
    ),
    overwrite: bool = Option(
        False,
        "--overwrite/--no-overwrite",
        help="Overwrite contents of output directory if it exists",
    ),
    verbose: bool = Option(False, "--verbose", "-v"),
    _: bool = Option(False, "--version", callback=utils.version_callback),
) -> None:
    """
    Read the Signal directory and output attachments and chat to DEST directory.

    Example to list chats:

        sigexport --list-chats

    Example to export all to a directory:

        sigexport ~/outputdir
    """
    logging.verbose = verbose

    if not any((dest, list_chats)):
        secho(ctx.get_help())
        # secho("Error: Missing argument 'DEST'", fg=colors.RED)
        raise Exit(code=1)

    if source:
        source_dir = Path(source).expanduser().absolute()
    else:
        source_dir = utils.source_location()
    if not (source_dir / "config.json").is_file():
        secho(f"Error: config.json not found in directory {source_dir}")
        raise Exit(code=1)

    convos, contacts = data.fetch_data(
        source_dir,
        password=password,
        key=key,
        chats=chats,
        include_empty=include_empty,
    )

    if list_chats:
        names = sorted(v.name for v in contacts.values() if v.name is not None)
        secho(" | ".join(names))
        raise Exit()

    dest = Path(dest).expanduser()
    if not dest.is_dir():
        dest.mkdir(parents=True, exist_ok=True)
    elif overwrite:
        shutil.rmtree(dest)
        dest.mkdir(parents=True, exist_ok=True)
    else:
        secho(
            f"Output folder '{dest}' already exists, didn't do anything!", fg=colors.RED
        )
        raise Exit()

    contacts = utils.fix_names(contacts)

    secho("Copying and renaming attachments")
    files.copy_attachments(source_dir, dest, convos, contacts)

    if json_output and old:
        secho(
            "Warning: currently, JSON does not support merging with the --old flag",
            fg=colors.RED,
        )

    secho("Creating output files")
    chat_dict = create.create_chats(convos, contacts)

    if old:
        secho(f"Merging old at {old} into output directory")
        secho("No existing files will be deleted or overwritten!")
        chat_dict = merge.merge_with_old(chat_dict, contacts, dest, Path(old))

    if paginate <= 0:
        paginate = int(1e20)

    if html_output:
        html.prep_html(dest)
    for key, messages in chat_dict.items():
        name = contacts[key].name
        # some contact names are None
        if not name:
            name = "None"
        md_path = dest / name / "chat.md"
        js_path = dest / name / "data.json"
        ht_path = dest / name / "index.html"

        md_f = md_path.open("a", encoding="utf-8")
        js_f = js_path.open("a", encoding="utf-8")
        ht_f = None
        if html_output:
            ht_f = ht_path.open("w", encoding="utf-8")

        try:
            for msg in messages:
                print(msg.to_md(), file=md_f)
                print(msg.dict_str(), file=js_f)
            if html_output:
                ht = html.create_html(
                    name=name, messages=messages, msgs_per_page=paginate
                )
                print(ht, file=ht_f)
        finally:
            md_f.close()
            js_f.close()
            if ht_f:
                ht_f.close()

    secho("Done!", fg=colors.GREEN)


def cli() -> None:
    """cli."""
    run(main)



"""Extract data from Signal DB."""

import json
from pathlib import Path
from typing import Optional

from sqlcipher3 import dbapi2
from typer import Exit, colors, secho

from sigexport import crypto, models
from sigexport.logging import log


def fetch_data(
    source_dir: Path,
    password: Optional[str],
    key: Optional[str],
    chats: str,
    include_empty: bool,
) -> tuple[models.Convos, models.Contacts]:
    """Load SQLite data into dicts."""
    db_file = source_dir / "sql" / "db.sqlite"

    if key is None:
        try:
            key = crypto.get_key(source_dir, password)
        except Exception as e:
            secho(f"Failed to decrypt Signal password: {e}", fg=colors.RED)
            raise Exit(1)

    log(f"Fetching data from {db_file}\n")
    contacts: models.Contacts = {}
    convos: models.Convos = {}
    chats_list = chats.split(",") if len(chats) > 0 else []

    db = dbapi2.connect(str(db_file))
    c = db.cursor()
    # param binding doesn't work for pragmas, so use a direct string concat
    c.execute(f"PRAGMA KEY = \"x'{key}'\"")
    c.execute("PRAGMA cipher_page_size = 4096")
    c.execute("PRAGMA kdf_iter = 64000")
    c.execute("PRAGMA cipher_hmac_algorithm = HMAC_SHA512")
    c.execute("PRAGMA cipher_kdf_algorithm = PBKDF2_HMAC_SHA512")

    query = "SELECT type, id, serviceId, e164, name, profileName, members FROM conversations"
    c.execute(query)
    for result in c:
        log(f"\tLoading SQL results for: {result[4]}, aka {result[5]}")
        members = []
        if result[6]:
            members = result[6].split(" ")
        is_group = result[0] == "group"
        cid = result[1]
        contacts[cid] = models.Contact(
            id=cid,
            serviceId=result[2],
            name=result[4],
            number=result[3],
            profile_name=result[5],
            members=members,
            is_group=is_group,
        )
        if contacts[cid].name is None:
            contacts[cid].name = contacts[cid].profile_name

        if not chats or (result[4] in chats_list or result[5] in chats_list):
            convos[cid] = []

    query = "SELECT json, conversationId FROM messages ORDER BY sent_at"
    c.execute(query)
    for result in c:
        res = json.loads(result[0])
        cid = result[1]
        if cid and cid in convos:
            if res.get("type") in ["keychange", "profile-change"]:
                continue
            con = models.RawMessage(
                conversation_id=res["conversationId"],
                id=res["id"],
                type=res.get("type"),
                body=res.get("body", ""),
                contact=res.get("contact"),
                source=res.get("sourceServiceId"),
                timestamp=res.get("timestamp"),
                sent_at=res.get("sent_at"),
                server_timestamp=res.get("serverTimestamp"),
                has_attachments=res.get("has_attachments", False),
                attachments=res.get("attachments", []),
                read_status=res.get("read_status"),
                seen_status=res.get("seen_status"),
                call_history=res.get("call_history"),
                reactions=res.get("reactions", []),
                sticker=res.get("sticker"),
                quote=res.get("quote"),
            )
            convos[cid].append(con)

    if not include_empty:
        convos = {key: val for key, val in convos.items() if len(val) > 0}

    return convos, contacts
